/*
 * pbutton.h
 *
 * Push button control ancestor class.
 *
 * Portable Windows Library
 *
 * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Portable Windows Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
 * All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: pbutton.h,v $
 * Revision 1.20  2001/05/22 12:49:33  robertj
 * Did some seriously wierd rewrite of platform headers to eliminate the
 *   stupid GNU compiler warning about braces not matching.
 *
 * Revision 1.19  1999/03/10 03:49:52  robertj
 * More documentation adjustments.
 *
 * Revision 1.18  1999/03/09 08:01:49  robertj
 * Changed comments for doc++ support (more to come).
 *
 * Revision 1.17  1999/02/16 08:08:46  robertj
 * MSVC 6.0 compatibility changes.
 *
 * Revision 1.16  1998/09/23 06:24:36  robertj
 * Added open source copyright license.
 *
 * Revision 1.15  1995/06/17 11:12:50  robertj
 * Documentation update.
 *
 * Revision 1.14  1995/03/14 12:41:58  robertj
 * Updated documentation to use HTML codes.
 *
 * Revision 1.13  1994/12/21  11:53:19  robertj
 * Documentation and variable normalisation.
 *
 * Revision 1.12  1994/10/30  11:46:56  robertj
 * Changed mechanism for doing notification callback functions.
 *
 * Revision 1.11  1994/08/23  11:32:52  robertj
 * Oops
 *
 * Revision 1.10  1994/08/22  00:46:48  robertj
 * Added pragma fro GNU C++ compiler.
 *
 * Revision 1.9  1994/04/03  08:34:18  robertj
 * Added help and focus functionality.
 *
 * Revision 1.8  1994/03/07  07:38:19  robertj
 * Major enhancementsacross the board.
 *
 * Revision 1.7  1994/01/03  04:42:23  robertj
 * Mass changes to common container classes and interactors etc etc etc.
 *
 * Revision 1.6  1993/09/27  16:35:25  robertj
 * Removed special constructor for dialog resource loading.
 *
 * Revision 1.5  1993/08/24  00:27:59  robertj
 * New push button class arrangement and callback function.
 *
 * Revision 1.4  1993/08/21  01:50:33  robertj
 * Made Clone() function optional, default will assert if called.
 *
 * Revision 1.3  1993/07/14  12:49:16  robertj
 * Fixed RCS keywords.
 *
 */


#define _PPUSHBUTTON

#ifdef __GNUC__
#pragma interface
#endif


/**A simple push button control to initiate an action.

   This is an abstract class for user defined types of push buttons, as well as
   the standard text and image buttons. When making a customised push button
   the application should not override the \Ref{PInteractor::OnRedraw()}
   function.

   The exact appearence of a push button is defined by the descendent class.
   However, they all have the same semantics of calling the notification
   function when they are selected, be that via mouse click or some other
   mechanism as is possible on the specific platform.
 */
class PPushButton : public PNamedControl
{
  PCLASSINFO(PPushButton, PNamedControl);

  public:
    /** Type of button for construction */
    enum ButtonType {
      /// Button appears as non-default buttons.
      OrdinaryButton,   
     /**Button appears as the default button, ie with some form of
         highlighting to indicate this.
       */
      DefaultButton
    };

      /** Create a push button control with the specified name. */
    PPushButton(
      PInteractor * parent,     /// Interactor into which the control is placed.
      ButtonType deflt = OrdinaryButton  /// Flag for is the default button.
    );
    PPushButton(
      PInteractor * parent,     /// Interactor into which the control is placed.
      const PString & name,     /// Name to be placed next in the button.
      ButtonType deflt,         /// Flag for is the default button.
      const PNotifier & notify  /// Function to call when changes state.
    );

   /**Create control from interactor layout with the specified control ID.
       This is mainly used in support of resource based dialogs.
     */
    PPushButton(
      PInteractorLayout * parent, /// Interactor into which the box is placed.
      PRESOURCE_ID ctlID,         /// Identifier for the control in the layout.
      const PNotifier & notify,   /// Function to call when changes state.
      void * valuePtr             /// Variable to change to the button state.
    );

    /** Destroy the push button control. */
    virtual ~PPushButton();


  /**@name New functions for class */
   /**Draw the 3D look of a pushbutton surround. The return value is passed
       to \Ref{OnDrawFace()} function as the bounds into which
       to draw.
       
       The default implementation draws a "lighted" rectangle that is raised
       when #isSelected# is FALSE and sunken when
       #isSelected# is TRUE. It will also draw a thicker border if
       the button is a default button.
    
       This function should be overridden when the button surround needs to be
       something other than the default. Do {\bf not} override the
       \Ref{PInteractor::OnRedraw()} function, override
       \Ref{DrawButtonSurround()} and/or the \Ref{OnDrawFace()} functions for
       drawing custom buttons. This is required by the various platforms
       implementations.

       @return
       rectangle for the "inside" of the button.
     */
    virtual PRect DrawButtonSurround(
      PCanvas & canvas,  /// Canvas into which to draw the button surround.
      BOOL isSelected    /// Flag for whether button is currently down.
    );


  /**@name System callback functions. */
   /**Draw the contents (face) of the push button.
    
       This is a pure function and must be overridden by descendent classes to
       give the button its visual appearance. It is called immediately after
       the \Ref{DrawButtonSurround()} function whenever the button needs to
       be redrawn.
       
       Note that due to the implementation just overriding the
       \Ref{PInteractor::OnRedraw()} function is insufficient to correctly
       implement a new look for a button. Always override the
       \Ref{DrawButtonSurround()} and \Ref{OnDrawFace()} functions.
     */
    virtual void OnDrawFace(
      PCanvas & canvas,   /// Canvas into which to draw the button face.
      const PRect & rect, /// Bounds in which to draw the button face.
      BOOL hasFocus,      /// Flag indicating that the button has the focus.
      BOOL isSelected     /// Flag indicating the button is currently down.
    ) = 0;


  protected:
    // New functions for class
    void Construct();
      // Common, platform specific, construction code.


    // Member variables
      /** Flag for button is a default button with extra highlighting */
    BOOL defaultButton;


// Include platform dependent part of class
#include <pwlib/pbutton.h>
};


// End Of File ///////////////////////////////////////////////////////////////
